#include "settings.h"
#include "strtoken.h"
#include <math.h>	//pow
#define _GDIPP_DLL
#include "supinfo.h"

static const TCHAR c_szGeneral[]  = _T("General");
static const TCHAR c_szFreeType[] = _T("FreeType");
#define HINTING_MIN		0
#define HINTING_MAX		2
#define AAMODE_MIN		-1
#define AAMODE_MAX		3
#define NWEIGHT_MIN		-64
#define NWEIGHT_MAX		+64
#define BWEIGHT_MIN		-32
#define BWEIGHT_MAX		+32
#define SLANT_MIN		-32
#define SLANT_MAX		+32

bool CGdiPPSettings::LoadSettings()
{
	CCriticalSectionLock __lock;
	Assert(m_szFileName[0]);
	return LoadAppSettings(m_szFileName);
}

bool CGdiPPSettings::LoadSettings(HINSTANCE hModule)
{
	CCriticalSectionLock __lock;
	if (!::GetModuleFileName(hModule, m_szFileName, MAX_PATH - sizeof(".ini") + 1)) {
		return false;
	}

	_tcscpy(PathFindExtension(m_szFileName), _T(".ini"));
	return LoadAppSettings(m_szFileName);
}

int CGdiPPSettings::_GetFreeTypeProfileInt(LPCTSTR lpszKey, int nDefault, LPCTSTR lpszFile)
{
	const int retA = GetPrivateProfileInt(c_szFreeType, lpszKey, -1, lpszFile);
	const int retB = GetPrivateProfileInt(c_szFreeType, lpszKey, -2, lpszFile);

	if (retA == retB) {
		return retA;
	}
	return GetPrivateProfileInt(c_szGeneral, lpszKey, nDefault, lpszFile);
}

int CGdiPPSettings::_GetFreeTypeProfileBoundInt(LPCTSTR lpszKey, int nDefault, int nMin, int nMax, LPCTSTR lpszFile)
{
	const int ret = _GetFreeTypeProfileInt(lpszKey, nDefault, lpszFile);
	return Bound(ret, nMin, nMax);
}

DWORD CGdiPPSettings::_GetFreeTypeProfileString(LPCTSTR lpszKey, LPCTSTR lpszDefault, LPTSTR lpszRet, DWORD cch, LPCTSTR lpszFile)
{
	const int retA = GetPrivateProfileInt(c_szFreeType, lpszKey, -1, lpszFile);
	const int retB = GetPrivateProfileInt(c_szFreeType, lpszKey, -2, lpszFile);

	if (retA == retB) {
		return GetPrivateProfileString(c_szFreeType, lpszKey, lpszDefault, lpszRet, cch, lpszFile);
	}
	return GetPrivateProfileString(c_szGeneral, lpszKey, lpszDefault, lpszRet, cch, lpszFile);
}

bool CGdiPPSettings::LoadAppSettings(LPCTSTR lpszFile)
{
	// eݒǂݍ
	// INIt@C̗:
	// [General]
	// HookChildProcesses=0
	// HintingMode=0
	// AntiAliasMode=0
	// NormalWeight=0
	// BoldWeight=0
	// ItalicSlant=0
	// EnableKerning=0
	// MaxHeight=0
	// ForceChangeFont=lr oSVbN
	// TextTuning=0
	// TextTuningR=0
	// TextTuningG=0
	// TextTuningB=0
	// CacheMaxFaces=0
	// CacheMaxSizes=0
	// CacheMaxBytes=0
	// AlternativeFile=
	// LoadOnDemand=0
	// UseMapping=0
	// LcdFilter=0
	// Shadow=1,1,4
	// [Individual]
	// lr oSVbN=0,1,2,3,4,5

	WritePrivateProfileString(NULL, NULL, NULL, lpszFile);

	TCHAR szAlternative[MAX_PATH];
	if (GetPrivateProfileString(c_szGeneral, _T("AlternativeFile"), _T(""), szAlternative, MAX_PATH, lpszFile)) {
		if (PathIsRelative(szAlternative)) {
			TCHAR szDir[MAX_PATH];
			lstrcpyn(szDir, lpszFile, MAX_PATH);
			PathRemoveFileSpec(szDir);
			PathCombine(szAlternative, szDir, szAlternative);
		}
		lstrcpyn(m_szFileName, szAlternative, MAX_PATH);
		lpszFile = m_szFileName;
		WritePrivateProfileString(NULL, NULL, NULL, lpszFile);
	}

	CFontSettings& fs = m_FontSettings;
	fs.Clear();
	fs.SetHintingMode(_GetFreeTypeProfileBoundInt(_T("HintingMode"), 0, HINTING_MIN, HINTING_MAX, lpszFile));
	fs.SetAntiAliasMode(_GetFreeTypeProfileBoundInt(_T("AntiAliasMode"), 0, AAMODE_MIN, AAMODE_MAX, lpszFile));
	fs.SetNormalWeight(_GetFreeTypeProfileBoundInt(_T("NormalWeight"), 0, NWEIGHT_MIN, NWEIGHT_MAX, lpszFile));
	fs.SetBoldWeight(_GetFreeTypeProfileBoundInt(_T("BoldWeight"), 0, BWEIGHT_MIN, BWEIGHT_MAX, lpszFile));
	fs.SetItalicSlant(_GetFreeTypeProfileBoundInt(_T("ItalicSlant"), 0, SLANT_MIN, SLANT_MAX, lpszFile));
	fs.SetKerning(!!_GetFreeTypeProfileInt(_T("EnableKerning"), 0, lpszFile));

	{
		TCHAR szShadow[256];
		CStringTokenizer token;
		m_bEnableShadow = false;
		if (!_GetFreeTypeProfileString(_T("Shadow"), _T(""), szShadow, countof(szShadow), lpszFile)
				|| token.Parse(szShadow) != 3) {
			goto SKIP;
		}
		for (int i=0; i<3; i++) {
			m_nShadow[i] = _StrToInt(token.GetArgument(i), 0);
			if (m_nShadow[i] <= 0) {
				goto SKIP;
			}
		}
		m_bEnableShadow = true;

SKIP:
		;
	}

	m_bHookChildProcesses = !!GetPrivateProfileInt(c_szGeneral, _T("HookChildProcesses"), false, lpszFile);
	m_bUseMapping	= !!_GetFreeTypeProfileInt(_T("UseMapping"), false, lpszFile);
	m_nBolderMode	= _GetFreeTypeProfileInt(_T("BolderMode"), 0, lpszFile);
	m_nGammaMode	= _GetFreeTypeProfileInt(_T("GammaMode"), 0, lpszFile);
//	m_dGammaValue	= _GetFreeTypeProfileDouble(_T("GammaValue"), 0, lpszFile);
	m_dGammaValue	= 2.2;
	m_bLoadOnDemand	= !!_GetFreeTypeProfileInt(_T("LoadOnDemand"), false, lpszFile);
	m_bIsInclude	= !!_GetFreeTypeProfileInt(_T("UseInclude"), false, lpszFile);
	m_nMaxHeight	= _GetFreeTypeProfileBoundInt(_T("MaxHeight"), 0, 0, 0x7fffffff, lpszFile);
	m_nLcdFilter	= _GetFreeTypeProfileInt(_T("LcdFilter"), 0, lpszFile);

	m_nCacheMaxFaces = _GetFreeTypeProfileInt(_T("CacheMaxFaces"), 0, lpszFile);
	m_nCacheMaxSizes = _GetFreeTypeProfileInt(_T("CacheMaxSizes"), 0, lpszFile);
	m_nCacheMaxBytes = _GetFreeTypeProfileInt(_T("CacheMaxBytes"), 0, lpszFile);

	// tHgw
	ZeroMemory(&m_lfForceFont, sizeof(LOGFONT));
	TCHAR szForceFont[LF_FACESIZE];
	szForceFont[0] = _T('\0');
	_GetFreeTypeProfileString(_T("ForceChangeFont"), _T(""), szForceFont, LF_FACESIZE, lpszFile);
	if (szForceFont[0]) {
		HDC hdc = GetDC(NULL);
		EnumFontFamilies(hdc, szForceFont, EnumFontFamProc, reinterpret_cast<LPARAM>(this));
		ReleaseDC(NULL, hdc);
	}

	const int nTextTuning	= _GetFreeTypeProfileInt(_T("TextTuning"),  0, lpszFile),
			  nTextTuningR	= _GetFreeTypeProfileInt(_T("TextTuningR"), 0, lpszFile),
			  nTextTuningG	= _GetFreeTypeProfileInt(_T("TextTuningG"), 0, lpszFile),
			  nTextTuningB	= _GetFreeTypeProfileInt(_T("TextTuningB"), 0, lpszFile);

	InitInitTuneTable();
	InitTuneTable(nTextTuning,  m_nTuneTable);
	InitTuneTable(nTextTuningR, m_nTuneTableR);
	InitTuneTable(nTextTuningG, m_nTuneTableG);
	InitTuneTable(nTextTuningB, m_nTuneTableB);

	// OS̃o[WXPȍ~ǂ
	OSVERSIONINFO osvi = { sizeof(OSVERSIONINFO) };
	GetVersionEx(&osvi);
	m_bIsWinXPorLater = ((osvi.dwMajorVersion > 5) || ((osvi.dwMajorVersion == 5) && (osvi.dwMinorVersion >= 1)));

	STARTUPINFO si = { sizeof(STARTUPINFO) };
	GetStartupInfo(&si);
	m_bRunFromGdiExe = IsGdiPPStartupInfo(si);
//	if (!m_bRunFromGdiExe) {
//		m_bHookChildProcesses = false;
//	}

//	m_bIsHDBench = (GetModuleHandle(_T("HDBENCH.EXE")) == GetModuleHandle(NULL));

/*	//CGdiPPSettings::LoadAppSettingsĂяoɂ͂܂FreeTypeĂȂ

	// FreeType2.2.2ǂ
	extern DWORD FreeTypeGetVersion();
	const DWORD ver = FreeTypeGetVersion();
	const BYTE major = GetRValue(ver);
	const BYTE minor = GetGValue(ver);
	const BYTE patch = GetBValue(ver);
//	m_bHaveNewerFreeType = (major > 2 || (major >= 2 && (minor > 2 || (minor >= 2 && patch >= 2))));
	if (major > 2)
		m_bHaveNewerFreeType = true;
	else if (major >= 2 && minor > 2)
		m_bHaveNewerFreeType = true;
	else if (major >= 2 && minor >= 2 && patch >= 2)
		m_bHaveNewerFreeType = true;
	else
		m_bHaveNewerFreeType = false;*/

	m_arrExcludeFont.RemoveAll();
	m_arrExcludeModule.RemoveAll();
	m_arrIncludeModule.RemoveAll();

	// [Exclude]ZNV珜OtHgXgǂݍ
	AddListFromSection(_T("Exclude"), lpszFile, m_arrExcludeFont);
	// [ExcludeModule]ZNV珜OW[Xgǂݍ
	AddListFromSection(_T("ExcludeModule"), lpszFile, m_arrExcludeModule);
	// [IncludeModule]ZNVΏۃW[Xgǂݍ
	AddListFromSection(_T("IncludeModule"), lpszFile, m_arrIncludeModule);

	// [Individual]ZNVtHgʐݒǂݍ
	AddIndividualFromSection(_T("Individual"), lpszFile, m_arrIndividual);
	return true;
}

int CALLBACK CGdiPPSettings::EnumFontFamProc(const LOGFONT* lplf, const TEXTMETRIC* lptm, DWORD FontType, LPARAM lParam)
{
	CGdiPPSettings* pThis = reinterpret_cast<CGdiPPSettings*>(lParam);
	if (pThis && FontType == TRUETYPE_FONTTYPE)
		pThis->m_lfForceFont = *lplf;
	return 0;
}

template <class T>
bool CGdiPPSettings::AddListFromSection(LPCTSTR lpszSection, LPCTSTR lpszFile, CArray<T>& arr)
{
	LPTSTR  buffer = _GetPrivateProfileSection(lpszSection, lpszFile);
	if (buffer == NULL) {
		SetLastError(ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	LPTSTR p = buffer;
	while (*p) {
		bool b = false;
		T str(p);
		for (int i = 0 ; i < arr.GetSize(); i++) {
			if (arr[i] == str) {
				b = true;
				break;
			}
		}
		if (!b) {
			arr.Add(str);
		}

		for (; *p; p++);
		p++;
	}
	delete[] buffer;
	return false;
}

bool CGdiPPSettings::AddIndividualFromSection(LPCTSTR lpszSection, LPCTSTR lpszFile, IndividualArray& arr)
{
	LPTSTR  buffer = _GetPrivateProfileSection(lpszSection, lpszFile);
	if (buffer == NULL) {
		SetLastError(ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	LPTSTR p = buffer;
	while (*p) {
		bool b = false;

		LPTSTR pnext = p;
		for (; *pnext; pnext++);

		//"lr oSVbN=0,0" ݂ȕ𕪊
		LPTSTR value = _tcschr(p, _T('='));
		CStringTokenizer token;
		int argc = 0;
		if (value) {
			*value++ = _T('\0');
			argc = token.Parse(value);
		}

		CFontIndividual fi(p);
		const CFontSettings& fsCommon = m_FontSettings;
		CFontSettings& fs = fi.GetIndividual();
		//Individual΋ʐݒg
		fs = fsCommon;
		for (int i = 0; i < MAX_FONT_SETTINGS; i++) {
			LPCTSTR arg = token.GetArgument(i);
			if (!arg)
				break;
			const int n = _StrToInt(arg, fsCommon.GetParam(i));
			fs.SetParam(i, n);
		}

		for (int i = 0 ; i < arr.GetSize(); i++) {
			if (arr[i] == fi) {
				b = true;
				break;
			}
		}
		if (!b) {
			arr.Add(fi);
#ifdef _DEBUG
			TRACE(_T("Individual: %s, %d, %d, %d, %d, %d, %d\n"), fi.GetName(),
					fs.GetParam(0), fs.GetParam(1), fs.GetParam(2), fs.GetParam(3), fs.GetParam(4), fs.GetParam(5));
#endif
		}

		p = pnext;
		p++;
	}
	delete[] buffer;
	return false;
}

LPTSTR CGdiPPSettings::_GetPrivateProfileSection(LPCTSTR lpszSection, LPCTSTR lpszFile)
{
	LPTSTR buffer = NULL;
	int nRes = 0;
	for (int cch = 256; ; cch += 256) {
		buffer = new TCHAR[cch];
		if (!buffer) {
			break;
		}

		ZeroMemory(buffer, sizeof(TCHAR) * cch);
		nRes = GetPrivateProfileSection(lpszSection, buffer, cch, lpszFile);
		if (nRes < cch - 2) {
			break;
		}
		delete[] buffer;
		buffer = NULL;
	}
	return buffer;
}

//atolɃftHglԂ悤ɂ悤ȕ
int CGdiPPSettings::_StrToInt(LPCTSTR pStr, int nDefault)
{
#define isspace(ch)		(ch == _T('\t') || ch == _T(' '))
#define isdigit(ch)		((_TUCHAR)(ch - _T('0')) <= 9)

	int ret;
	bool neg = false;
	LPCTSTR pStart;

	for (; isspace(*pStr); pStr++);
	switch (*pStr) {
	case _T('-'):
		neg = true;
	case _T('+'):
		pStr++;
		break;
	}

	pStart = pStr;
	ret = 0;
	for (; isdigit(*pStr); pStr++) {
		ret = 10 * ret + (*pStr - _T('0'));
	}

	if (pStr == pStart) {
		return nDefault;
	}
	return neg ? -ret : ret;

#undef isspace
#undef isdigit
}

bool CGdiPPSettings::IsFontExcluded(LPCSTR lpFaceName) const
{
	WCHAR szStack[LF_FACESIZE];
	LPWSTR lpUnicode = _StrDupExAtoW(lpFaceName, -1, szStack, LF_FACESIZE, NULL);
	if (!lpUnicode) {
		return false;
	}

	bool b = IsFontExcluded(lpUnicode);
	if (lpUnicode != szStack)
		free(lpUnicode);
	return b;
}

bool CGdiPPSettings::IsFontExcluded(LPCWSTR lpFaceName) const
{
	StringHashFont* p = m_arrExcludeFont.Begin();
	StringHashFont* end = m_arrExcludeFont.End();
	StringHashFont str(lpFaceName);

	for(; p != end; ++p) {
		if (*p == str) {
			return true;
		}
	}
	return false;
}

bool CGdiPPSettings::IsProcessExcluded() const
{
	if (m_bRunFromGdiExe) {
		return false;
	}
	StringHashModule* p = m_arrExcludeModule.Begin();
	StringHashModule* end = m_arrExcludeModule.End();
	for(; p != end; ++p) {
		if (GetModuleHandleW(p->c_str())) {
			return true;
		}
	}
	return false;
}

bool CGdiPPSettings::IsProcessIncluded() const
{
	if (m_bRunFromGdiExe) {
		return true;
	}
	StringHashModule* p = m_arrIncludeModule.Begin();
	StringHashModule* end = m_arrIncludeModule.End();
	for(; p != end; ++p) {
		if (GetModuleHandleW(p->c_str())) {
			return true;
		}
	}
	return false;
}

void CGdiPPSettings::InitInitTuneTable()
{
	int i, *table;
#define init_table(name) \
		for (i=0,table=name; i<256; i++) table[i] = i
	init_table(m_nTuneTable);
	init_table(m_nTuneTableR);
	init_table(m_nTuneTableG);
	init_table(m_nTuneTableB);
#undef init_table
}

// e[u֐ 0 - 12܂
// LCDpe[u֐ e0 - 12܂
void CGdiPPSettings::InitTuneTable(int v, int* table)
{
	int i;
	int col;
	double tmp, p;

	if (v < 0) {
		return;
	}
	v = Min(v, 12);
	p = (double)v;
	p = 1 - (p / (p + 10.0));
	for(i = 0;i < 256;i++){
	    tmp = (double)i / 255.0;
        tmp = pow(tmp, p);
	    col = 255 - (int)(tmp * 255.0 + 0.5);
		table[255 - i] = col;
	}
}

//Ȃꍇ͋ʐݒԂ
const CFontSettings& CGdiPPSettings::FindIndividual(LPCTSTR lpFaceName) const
{
	CFontIndividual* p		= m_arrIndividual.Begin();
	CFontIndividual* end	= m_arrIndividual.End();
	StringHashFont hash(lpFaceName);

	for(; p != end; ++p) {
		if (p->GetHash() == hash) {
			return p->GetIndividual();
		}
	}
	return GetFontSettings();
}

//lIchar(-128`127)ŏ\
const char CFontSettings::m_bound[MAX_FONT_SETTINGS][2] = {
	{ HINTING_MIN,	HINTING_MAX	},	//Hinting
	{ AAMODE_MIN,	AAMODE_MAX	},	//AAMode
	{ NWEIGHT_MIN,	NWEIGHT_MAX	},	//NormalWeight
	{ BWEIGHT_MIN,	BWEIGHT_MAX	},	//BoldWeight
	{ SLANT_MIN,	SLANT_MAX	},	//ItalicSlant
	{ 0,			1			},	//Kerning
};
